#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>
#include <multithreading.h>

//add n to all the elements of devarr.
__global__ void kAdd(float* devarr, float n) {
	//get the index in our array
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	devarr[idx] += n;
}

//the array we send to the device
float* hostarray;
//the array we get back
float* returnarray;
//array on the device
__device__ float* devarray;

int main(int argc, char **argv) {
	int GPU_N;
	int arrsize;
	int tpb;	//threads per block
	cudaDeviceProp dprop;

	dim3 threads;
	dim3 grid;

	int i;
	float eps = 1e-6f;
	float nAdd = 0.5f;
	int err;

	//CUDA_SAFE_CALL prints out debugging information if errors occur
	CUDA_SAFE_CALL(cudaGetDeviceCount(&GPU_N));
    printf("CUDA-capable device count: %i\n", GPU_N);
	if (GPU_N == 0) {
		exit(1);
	}
	CUT_DEVICE_INIT(argc, argv);

	//find out some information about the device
	CUDA_SAFE_CALL(cudaGetDeviceProperties(&dprop, 0));
	printf("Maximum threads per block: %d\n", dprop.maxThreadsPerBlock);

	tpb = dprop.maxThreadsPerBlock;
	//setup the thread and grid arrangement
	threads = dim3(tpb, 1, 1);
	arrsize = tpb*2;
	grid = dim3(arrsize/tpb, 1, 1);

	printf("The size of a float is %d bytes\n", sizeof(float));
	printf("Creating %d threads and a %d long list\n", tpb, arrsize);
	
	//create an array on the host
	hostarray = (float*) malloc(sizeof(float)*arrsize);
	returnarray = (float*) malloc(sizeof(float)*arrsize);
	memset(returnarray, 0, sizeof(float)*arrsize);

	for (i = 0; i < arrsize; ++i) {
		hostarray[i] = (float) (i % arrsize);
	}

	//allocate memory on the device
	CUDA_SAFE_CALL(cudaMalloc((void**)&devarray, sizeof(float)*arrsize));
	CUDA_SAFE_CALL(cudaMemset(devarray, 0, sizeof(float)*arrsize));

	//copy over the array
	CUDA_SAFE_CALL(cudaMemcpy(devarray, hostarray, sizeof(float)*arrsize, cudaMemcpyHostToDevice));

	//execute the kernel
	kAdd<<<grid, threads>>>(devarray, nAdd);
	CUT_CHECK_ERROR("Error running GPU kernel");

	//copy the array back
	CUDA_SAFE_CALL(cudaMemcpy(returnarray, devarray, sizeof(float)*arrsize, cudaMemcpyDeviceToHost));

	//clear the device memory
	CUDA_SAFE_CALL(cudaFree(devarray));

	//compare the arrays
	
	//print out the first few values
	printf("Comparing the first few entries...\n");
	printf("%8s -> %8s\n", "sent", "return");
	for (i = 0; i < 10; ++i) {
		printf("%8f -> %8f\n", hostarray[i], returnarray[i]);
	}

	//compare the whole array.
	err = 0;
	printf("Comparing all entries to eps=%f\n", eps);
	for (i = 0; i < arrsize; ++i) {
		if (abs(hostarray[i] + nAdd - returnarray[i]) > eps) {
			err = 1;
			printf("Error at index %d: %f -> %f\n", hostarray[i], returnarray[i]);
		}
	}
	if (!err) {
		printf("No errors encountered.\n");
	}

    CUT_EXIT(argc, argv);
}
