/*
GPU kernels and definitions for fractal computation
Ilari Shafer, 12/13/08
*/
typedef unsigned int uint;

//random number array dimensions
const int REPSMAX = 10000*8;
const int RANDNODIM = REPSMAX/8;

//the total length of the framebuffer
const int BUFFERLEN = 8;
const int RANDFLOATDIM = BUFFERLEN * 2;

//Maintain a string of random numbers in constant memory cache
//It's not as optimal as using texture memory, but it's much cleaner,
//and we don't need 2D access optimization.
__device__ __constant__ int randnos[RANDNODIM];
//Have a few random numbers as starting positions
__device__ __constant__ float randfloats[RANDFLOATDIM];
uint* output[BUFFERLEN];

//Do stupid things to get things into global memory.
//	Passing in double pointers as kernel parameters is badness (see old/chunkdraw.cu)
typedef struct {
	uint* out[BUFFERLEN];
} outpointer;

//Compute the fractal
__global__ void kdraw(outpointer op, GPUParams* params, int width, int height, int reps) {
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	float x,y,xn,yn;
	int xt,yt;
	int i, j;

	//copy from global to shared memory
	__shared__ GPUParams tp;
	tp = params[0];

	//Start at the middle for testing purposes
	//x = (tp.xmin + tp.xmax)/2;
	//y = (tp.ymin + tp.ymax)/2;

	//Start at a random location
	x = randfloats[idx*2] * (tp.xmax - tp.xmin) + tp.xmin;
	y = randfloats[idx*2+1] * (tp.ymax - tp.ymin) + tp.ymin;

	//Use different random numbers for each thread
	for (i = reps*idx; i < reps*idx+reps; ++i) {
		//This should already be done for us... but unroll the loop 32 times
		#pragma unroll 32
		for (j = 0; j < sizeof(int)*8; ++j) {
			//bitshift the mask over by one each time
			//this effectively checks one random 0/1 at a time
			if (randnos[i] & (1 << j)) {
				xn = tp.a1 * x + tp.c1 * y + tp.e1;
				yn = tp.b1 * x + tp.d1 * y + tp.f1;
			} else {
				xn = tp.a2 * x + tp.c2 * y + tp.e2;
				yn = tp.b2 * x + tp.d2 * y + tp.f2;
			}

			xt = (int) ((xn - tp.xmin) / (tp.xmax - tp.xmin) * width);
			yt = (int) ((yn - tp.ymin) / (tp.ymax - tp.ymin) * height);

			//Use output buffer 0
			if (xt < width && xt > 0 && yt < height && yt > 0) {
				//requires -arch sm_11
				//Since multiple threads are accessing the data element, we have to add atomically.
				//Note that 
				atomicAdd(&op.out[0][yt*width+xt], 0x1);

				//This causes some syncronization problems, evidenced by a bit of flicker.
				//op.out[0][yt*width + xt] += 0x44;
			}

			x = xn;
			y = yn;
		}
	}
}

//Draw out details by executing a colormap operation on the image
__global__ void mapcolors(outpointer op, int width, int height) {
	int x = blockIdx.x * blockDim.x + threadIdx.x;
	int y = blockIdx.y * blockDim.y + threadIdx.y;

	//for demonstrating what happens without a ^(1/4) map - this is a straight map
	//op.out[0][y*width+x] = min(254, op.out[0][y*width+x] * 40);
	
	//Use a ^(1/4) map
	op.out[0][y*width+x] = (int)(sqrt(sqrt((float) op.out[0][y*width+x] / 256.0))*256.0);
}