#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_gl_interop.h>
#include <multithreading.h>

#include "util.h"

typedef unsigned int uint;

//the number of repetitions to run
const int REPS = 10000*8;
const int RANDNODIM = REPS/8;

//the total length of the framebuffer
const int BUFFERLEN = 32*2;

const int width = 512, height = 512;

//Do stupid things to get things into global memory.
//	Passing in double pointers as kernel parameters is badness (see old/chunkdraw.cu)
typedef struct {
	uint* out[BUFFERLEN/2];
} outpointer;

//the dimensions of the block size can't multiply to be more than 512.
const dim3 blockSize(8, 1, 1);
const dim3 gridSize(4, 1);

//OpenGL pixel buffer
GLuint pbo[BUFFERLEN];

//Maintain a string of random numbers in constant memory cache
//It's not as optimal as using texture memory, but it's much cleaner,
//and we don't need 2D access optimization.
__device__ __constant__ int randnos[RANDNODIM];
uint* output[BUFFERLEN];

__global__ void kdraw(outpointer op, int width, int height) {
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	int x, y;

	for (x = idx*4; x < idx*5; ++x) {
		for (y = idx*4; y < idx*5; ++y) {
			op.out[idx][y*width + x] = 0xffffffff;
		}
	}
}

void displayF() {
    char title[256];
	int x;
    sprintf(title, "Fractal %s", "asdf");  
    glutSetWindowTitle(title);

	for (x = 0; x < BUFFERLEN/2; ++x) {
		CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&output[x], pbo[x]));
	}

	//load up the output pointer with the values that were just mapped.
	outpointer op;
	for (x = 0; x < BUFFERLEN/2; ++x) {
		op.out[x] = output[x];
	}

    //call CUDA kernel, writing results to PBO
    kdraw<<<gridSize, blockSize>>>(op, width, height);
    CUT_CHECK_ERROR("kernel failed");

	for (x = 0; x < BUFFERLEN/2; ++x) {
		CUDA_SAFE_CALL(cudaGLUnmapBufferObject(pbo[x]));
	}

    // display results
    glClear(GL_COLOR_BUFFER_BIT);

    // draw image from PBO
    glDisable(GL_DEPTH_TEST);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo[30]);

	//we could do a glEnable(GL_BLEND) to use source/destination alpha
    glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

    glutSwapBuffers();
    glutReportErrors();
}

void idleF() {
	//do while idle...
	glutPostRedisplay();
}

void cleanup() {
	int x;
	for (x = 0; x < BUFFERLEN; ++x) {
		CUDA_SAFE_CALL(cudaGLUnregisterBufferObject(pbo[x]));
	}
	glDeleteBuffersARB(BUFFERLEN, pbo);
}

void initOpenGL() {
	int x;
	
	//generate buffer identifiers
	glGenBuffersARB(BUFFERLEN, pbo);

	// create pixel buffer objects
	for (x = 0; x < BUFFERLEN; ++x) {
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo[x]);
		glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, width*height*sizeof(GLubyte)*4, 0, GL_STREAM_DRAW_ARB);
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
		CUDA_SAFE_CALL(cudaGLRegisterBufferObject(pbo[x]));
	}
}

void loadRandom() {
	int data[RANDNODIM];
	int x;
	int y;

	//clear the array we're passing over.
	memset(data, 0, sizeof(data));

	//fill the array with random bits
	for (x = 0; x < RANDNODIM; ++x) {
		for (y = 0; y < sizeof(int)*8; ++y) {
			data[x] |= ((RAND1() > 0.5) ? 1 : 0) << y;
		}
	}

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(randnos,data,sizeof(data)));
}

int main(int argc, char **argv){
	CUT_DEVICE_INIT(argc, argv);

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(width, height);
	glutCreateWindow("GPU Computation to Direct Display");
    glutDisplayFunc(displayF);
	glutIdleFunc(idleF);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0 GL_ARB_pixel_buffer_object")) {
        fprintf(stderr, "Required OpenGL pixel buffer extension missing.");
        exit(-1);
    }
    initOpenGL();

	loadRandom();

	atexit(cleanup);

	glutMainLoop();

    CUT_EXIT(argc, argv);
}
