#include "fractal.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_gl_interop.h>
#include <multithreading.h>
#include "util.h"
#include "paramfile.h"

//the number of repetitions to run
#define REPS 300

typedef unsigned int uint;

//random number array dimensions
const int REPSMAX = 10000*8;
const int RANDNODIM = REPSMAX/8;

//the total length of the framebuffer
const int BUFFERLEN = 8;
const int RANDFLOATDIM = BUFFERLEN * 2;

const int width = 512, height = 512;

//Do stupid things to get things into global memory.
//	Passing in double pointers as kernel parameters is badness (see old/chunkdraw.cu)
typedef struct {
	uint* out[BUFFERLEN];
} outpointer;

//For using just one data output array, it's much faster to parallelize 
//than to increase the number of iterations (via REPS)
//the dimensions of the block size can't multiply to be more than 512.
const dim3 blockSize(4);
//make sure the first dimension of the grid size is more than the number of multiprocessors.
const dim3 gridSize(8);

const char* fnames[10] = {
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a001.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a002.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a003.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a004.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a005.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a006.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a007.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a008.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a009.ifs.txt",
	"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a010.ifs.txt"
};

ParamList* plst;
ParamTween* ptween;

//OpenGL pixel buffer
GLuint pbo[BUFFERLEN];

//Maintain a string of random numbers in constant memory cache
//It's not as optimal as using texture memory, but it's much cleaner,
//and we don't need 2D access optimization.
__device__ __constant__ int randnos[RANDNODIM];
//Have a few random numbers as starting positions
__device__ __constant__ float randfloats[RANDFLOATDIM];
uint* output[BUFFERLEN];

__device__ GPUParams* devparams;
GPUParams* hostparams;

__global__ void kdraw(outpointer op, GPUParams* params, int width, int height) {
	int idx = blockIdx.x * blockDim.x + threadIdx.x;
	float x,y,xn,yn;
	int xt,yt;
	int i, j;

	//copy from global to shared memory
	__shared__ GPUParams tp;
	tp = params[0];

	//Start at the middle for testing purposes
	//x = (tp.xmin + tp.xmax)/2;
	//y = (tp.ymin + tp.ymax)/2;

	//Start at a random location
	x = randfloats[idx*2] * (tp.xmax - tp.xmin) + tp.xmin;
	y = randfloats[idx*2+1] * (tp.ymax - tp.ymin) + tp.ymin;

	//Use different random numbers for each thread
	for (i = REPS*idx; i < REPS*idx+REPS; ++i) {
		//This should already be done for us... but unroll the loop 8 times
		#pragma unroll 8
		for (j = 0; j < sizeof(int)*8; ++j) {
			//bitshift the mask over by one each time
			//this effectively checks one random 0/1 at a time
			if (randnos[i] & (1 << j)) {
				xn = tp.a1 * x + tp.c1 * y + tp.e1;
				yn = tp.b1 * x + tp.d1 * y + tp.f1;
			} else {
				xn = tp.a2 * x + tp.c2 * y + tp.e2;
				yn = tp.b2 * x + tp.d2 * y + tp.f2;
			}

			xt = (int) ((xn - tp.xmin) / (tp.xmax - tp.xmin) * width);
			yt = (int) ((yn - tp.ymin) / (tp.ymax - tp.ymin) * height);

			//Use output buffer 0
			if (xt < width && xt > 0 && yt < height && yt > 0) {
				//requires -arch sm_11
				//Since multiple threads are accessing the data element, we have to add atomically.
				//Note that 
				atomicAdd(&op.out[0][yt*width+xt], 0x1);

				//This causes some syncronization problems, evidenced by a bit of flicker.
				//op.out[0][yt*width + xt] += 0x44;
			}

			x = xn;
			y = yn;
		}
	}

	/*
	for (x = idx*4; x < idx*5; ++x) {
		for (y = idx*4; y < idx*5; ++y) {
			op.out[idx][y*width + x] = 0xffffffff;
		}
	}*/
}

void displayF() {
    char title[256];
	int x;
    sprintf(title, "Fractal %s", "asdf");  
    glutSetWindowTitle(title);

	//It is much faster to use one output buffer
	for (x = 0; x < 1; ++x) {
		//map the GL buffer object
		CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&output[x], pbo[x]));

		//clear the buffer data
		CUDA_SAFE_CALL(cudaMemset((void*)output[x], 0, sizeof(uint)*width*height));
	}

	//load up the output pointer with the values that were just mapped.
	outpointer op;
	for (x = 0; x < BUFFERLEN; ++x) {
		op.out[x] = output[x];
	}

	//find the appropriate parameter list
	hostparams[0] = getGPUParams(ParamTween_cur(ptween));
	ParamTween_increment(ptween);

    unsigned int hTimer;
    CUT_SAFE_CALL(cutCreateTimer(&hTimer));
    CUT_SAFE_CALL(cutResetTimer(hTimer));
    CUT_SAFE_CALL(cutStartTimer(hTimer));

	//send the parameter list to the device
	CUDA_SAFE_CALL(cudaMemcpy(devparams, hostparams, sizeof(GPUParams)*BUFFERLEN, cudaMemcpyHostToDevice));

    //call CUDA kernel, writing results to PBO
    kdraw<<<gridSize, blockSize>>>(op, devparams, width, height);
    CUT_CHECK_ERROR("kernel failed");

	for (x = 0; x < 1; ++x) {
		CUDA_SAFE_CALL(cudaGLUnmapBufferObject(pbo[x]));
	}

	printf("Time taken: %f ms\n", cutGetTimerValue(hTimer));
    CUT_SAFE_CALL(cutDeleteTimer(hTimer));

    // display results
    glClear(GL_COLOR_BUFFER_BIT);

    // draw image from PBO
    glDisable(GL_DEPTH_TEST);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo[0]);

	//we could do a glEnable(GL_BLEND) to use source/destination alpha
	//if we actually used multiple pixel unpack buffers
    glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

    glutSwapBuffers();
    glutReportErrors();

	//Increment the frame
}

void idleF() {
	//do while idle...
	glutPostRedisplay();
}

void cleanup() {
	int x;
	for (x = 0; x < 1; ++x) {
		CUDA_SAFE_CALL(cudaGLUnregisterBufferObject(pbo[x]));
	}
	glDeleteBuffersARB(BUFFERLEN, pbo);

	//clean up the parameter storage space
	CUDA_SAFE_CALL(cudaFree(devparams));
	free(hostparams);
}

void initOpenGL() {
	int x;
	
	//generate buffer identifiers
	glGenBuffersARB(BUFFERLEN, pbo);

	// create pixel buffer objects
	for (x = 0; x < 1; ++x) {
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo[x]);
		glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, width*height*sizeof(GLubyte)*4, 0, GL_DYNAMIC_DRAW_ARB);
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
		CUDA_SAFE_CALL(cudaGLRegisterBufferObject(pbo[x]));
	}
}

void loadRandom() {
	int data[RANDNODIM];
	float fldata[RANDFLOATDIM];
	int x;
	int y;

	//clear the array we're passing over.
	memset(data, 0, sizeof(data));

	//fill the array with random bits
	for (x = 0; x < RANDNODIM; ++x) {
		for (y = 0; y < sizeof(int)*8; ++y) {
			data[x] |= ((RAND1() > 0.5) ? 1 : 0) << y;
		}
	}

	//put some random numbers in the random float array
	for (x = 0; x < RANDFLOATDIM; ++x) {
		fldata[x] = (float) RAND1();
	}

	//copy random numbers to the device
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(randnos,data,sizeof(data)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(randfloats,fldata,sizeof(fldata)));
}

int main(int argc, char **argv){
	CUT_DEVICE_INIT(argc, argv);

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(width, height);
	glutCreateWindow("GPU Computation to Direct Display");
    glutDisplayFunc(displayF);
	glutIdleFunc(idleF);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0 GL_ARB_pixel_buffer_object")) {
        fprintf(stderr, "Required OpenGL pixel buffer extension missing.");
        exit(-1);
    }
    initOpenGL();

	loadRandom();

	atexit(cleanup);

	//Create parameter list support
	CUDA_SAFE_CALL(cudaMalloc((void**)&devparams, sizeof(GPUParams)*BUFFERLEN));
	hostparams = (GPUParams*) malloc(sizeof(GPUParams)*BUFFERLEN);

	//For a single parameter list
	/*
	plst = ParamList_init();
	ParamList_load(plst, 
		"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a001.ifs.txt");
	GPUParams_print(getGPUParams(plst));
	*/

	//For multiple parameter lists
	ptween = ParamTween_init();
	ParamTween_load(ptween, fnames, 10, 200);

	glutMainLoop();

    CUT_EXIT(argc, argv);
}
