#include "paramfile.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#ifdef __cplusplus
extern "C" {
#endif

//allocate and return an empty parameter list
ParamList* ParamList_init() {
	ParamList* plst = (ParamList*) malloc(sizeof(ParamList));
	return plst;
}

void Param_print(Param par) {
	printf("\tParam {a:%5g b:%5g c:%5g d:%5g e:%5g f:%5g p:%5g}\n",
		par.a, par.b, par.c, par.d, par.e, par.f, par.p);
}

//print a formatted version of the given parameter list
void ParamList_print(ParamList* plst) {
	int x = 0;
	printf("ParamList {\n");
	for (x = 0; x < plst->paramlen; ++x) {
		Param_print(plst->paramarr[x]);
	}
	printf("\txbnd: [%5g, %5g]\n", plst->xmin, plst->xmax);
	printf("\tybnd: [%5g, %5g]\n", plst->ymin, plst->ymax);

	printf("}\n");
}

GPUParams getGPUParams(ParamList* plst) {
	GPUParams parms;
	parms.a1 = (float) plst->paramarr[0].a;
	parms.b1 = (float) plst->paramarr[0].b;
	parms.c1 = (float) plst->paramarr[0].c;
	parms.d1 = (float) plst->paramarr[0].d;
	parms.e1 = (float) plst->paramarr[0].e;
	parms.f1 = (float) plst->paramarr[0].f;

	parms.a2 = (float) plst->paramarr[1].a;
	parms.b2 = (float) plst->paramarr[1].b;
	parms.c2 = (float) plst->paramarr[1].c;
	parms.d2 = (float) plst->paramarr[1].d;
	parms.e2 = (float) plst->paramarr[1].e;
	parms.f2 = (float) plst->paramarr[1].f;

	parms.xmin = (float) plst->xmin;
	parms.xmax = (float) plst->xmax;
	parms.ymin = (float) plst->ymin;
	parms.ymax = (float) plst->ymax;

	return parms;
}

void GPUParams_print(GPUParams parms) {
	printf("GPUParams {\n");
	printf("\ttfm1: [%6f %6f %6f %6f %6f %6f]\n", parms.a1, parms.b1, parms.c1, parms.d1, parms.e1, parms.f1);
	printf("\ttfm2: [%6f %6f %6f %6f %6f %6f]\n", parms.a2, parms.b2, parms.c2, parms.d2, parms.e2, parms.f2);
	printf("\txbnd: [%6f %6f]\n", parms.xmin, parms.xmax);
	printf("\tybnd: [%6f %6f]\n", parms.ymin, parms.ymax);
	printf("}\n");
}

//tween between parameter lists one and two, by fraction frac
//	dest is the destination ParamList (so that we don't have to allocate more space)
//	note that the sizes of the lists in dest, one, and two must be identical.
//	frac is between 0 and 1
void ParamList_tween(ParamList* dest, ParamList* one, ParamList* two, double frac) {
	double omf = 1 - frac;
	int x;
	Param* ppt;
	Param ppa, ppb;
	dest->xmin = one->xmin * omf + two->xmin * frac;
	dest->xmax = one->xmax * omf + two->xmax * frac;
	dest->ymin = one->ymin * omf + two->ymin * frac;
	dest->ymax = one->ymax * omf + two->ymax * frac;

	for (x = 0; x < one->paramlen; ++x) {
		ppt = &dest->paramarr[x];
		ppa = one->paramarr[x];
		ppb = two->paramarr[x];
		ppt->a = ppa.a * omf + ppb.a * frac;
		ppt->b = ppa.b * omf + ppb.b * frac;
		ppt->c = ppa.c * omf + ppb.c * frac;
		ppt->d = ppa.d * omf + ppb.d * frac;
		ppt->e = ppa.e * omf + ppb.e * frac;
		ppt->f = ppa.f * omf + ppb.f * frac;
		ppt->p = ppa.p * omf + ppb.p * frac;
	}
}

//Load a parameter list from a file
void ParamList_load(ParamList* plst, const char* fname) {
	FILE* fin = fopen(fname, "r");
	char* fchars = (char*) malloc(512);

	//HACK: allocate extra space for the parameter list
	plst->paramarr = (Param*) malloc(sizeof(Param)*MAX_PARAM_LEN);
	plst->paramlen = 0;


	while (!feof(fin)) {
		//scan the first digit
		fscanf(fin, "%s", fchars);

		//skip comments
		if (fchars[0] == '#') {
			fgets(fchars, 512, fin);
		} else {
			if (atoi(fchars) == 0) {
				//encountered the last line, which should contain a "window" parameter
				//format: xmin ymin xmax ymax
				fscanf(fin, "%lg%lg%lg%lg",
					&plst->ymin,
					&plst->xmin,
					&plst->ymax,
					&plst->xmax
					);
				break;
			} else {
				//first digit was a line number; parse the remaining 7 values
				fscanf(fin, "%lg%lg%lg%lg%lg%lg%lg", 
					&plst->paramarr[plst->paramlen].a,
					&plst->paramarr[plst->paramlen].b,
					&plst->paramarr[plst->paramlen].c,
					&plst->paramarr[plst->paramlen].d,
					&plst->paramarr[plst->paramlen].e,
					&plst->paramarr[plst->paramlen].f,
					&plst->paramarr[plst->paramlen].p
					);
				++plst->paramlen;
			}
		}
	}

	fclose(fin);
}

ParamTween* ParamTween_init() {
	ParamTween* ptw = (ParamTween*) malloc(sizeof(ParamTween));
	return ptw;
}

void ParamTween_load(ParamTween* ptw, const char** fnames, int fnameslen, int tweensteps) {
	int x;

	ptw->tweenarr = (ParamList**) malloc(sizeof(ParamList*)*fnameslen);
	for (x = 0; x < fnameslen; ++x) {
		ptw->tweenarr[x] = ParamList_init();
		ParamList_load(ptw->tweenarr[x], fnames[x]);
	}
	ptw->tweenlen = fnameslen;

	ptw->curstep = 0;
	ptw->tweensteps = tweensteps;

	//initialize the current step to the first one in the list
	ptw->buf = ParamList_init();
	ParamList_load(ptw->buf, fnames[0]);
}

ParamList* ParamTween_cur(ParamTween* ptw) {
	//find the target indices
	int from = (int) floor(ptw->curstep / ptw->tweensteps);
	int to = (from == ptw->tweenlen - 1) ? 0 : from + 1;
	double frac = ((double)(ptw->curstep % ptw->tweensteps)) / (double) ptw->tweensteps;

	//compute and return the tweened step
	ParamList_tween(ptw->buf, ptw->tweenarr[from], ptw->tweenarr[to], frac);
	return ptw->buf;
}

void ParamTween_increment(ParamTween* ptw) {
	++ptw->curstep;
	if (ptw->curstep >= ptw->tweenlen * ptw->tweensteps) {
		ptw->curstep = 0;
	}
}

#ifdef __cplusplus
}
#endif
