#include "fractal.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_gl_interop.h>
#include <multithreading.h>
#include "util.h"
#include "paramfile.h"
#include "fileio.h"

#include "kernels.cu"

const int width = 512, height = 512;

//Colormap size
const int COLORMAPSIZE = 256;

ParamList* plst;
ParamTween* ptween;

//OpenGL pixel buffer
GLuint pbo[BUFFERLEN];

//For computing the reference solution on the CPU
int* cpudata;

__device__ GPUParams* devparams;
GPUParams* hostparams;

//Return four times:
//	Init
//	Draw
//	Map
//	Finish
float* GPUFrame(int iterations, unsigned int hTimer) {
	int x;
	float* times = (float*) malloc(sizeof(float)*4);
    CUT_SAFE_CALL(cutResetTimer(hTimer));
    CUT_SAFE_CALL(cutStartTimer(hTimer));

	//It is much faster to use one output buffer
	for (x = 0; x < 1; ++x) {
		//map the GL buffer object
		CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&output[x], pbo[x]));

		//clear the buffer data
		CUDA_SAFE_CALL(cudaMemset((void*)output[x], 0, sizeof(uint)*width*height));
	}

	//load up the output pointer with the values that were just mapped.
	outpointer op;
	for (x = 0; x < BUFFERLEN; ++x) {
		op.out[x] = output[x];
	}

	//find the appropriate parameter list
	hostparams[0] = getGPUParams(ParamTween_cur(ptween));
	ParamTween_increment(ptween);

	//send the parameter list to the device
	CUDA_SAFE_CALL(cudaMemcpy(devparams, hostparams, sizeof(GPUParams)*BUFFERLEN, cudaMemcpyHostToDevice));

	times[0] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

	//For using just one data output array, it's much faster to parallelize 
	//than to increase the number of iterations (via REPS)
	//the dimensions of the block size can't multiply to be more than 512.
	const dim3 drawBlockSize(4);
	//make sure the first dimension of the grid size is more than the number of multiprocessors.
	const dim3 drawGridSize(8);
    //call CUDA kernel, writing results to PBO
	//	There are sizeof(int)*8 sub-iterations per iteration
    kdraw<<<drawGridSize, drawBlockSize>>>(op, devparams, width, height, (int) (iterations/(sizeof(int)*8)/drawBlockSize.x/drawGridSize.x));
    CUT_CHECK_ERROR("Drawing kernel failed");

	//Make sure we finished drawing before colormapping
	cudaThreadSynchronize();

	times[1] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

	//Perform colormappping
	const dim3 mapBlockSize(16, 16, 1);
	const dim3 mapGridSize(width/mapBlockSize.x, height/mapBlockSize.y, 1);
	mapcolors<<<mapGridSize, mapBlockSize>>>(op, width, height);
	CUT_CHECK_ERROR("Colormap kernel failed");

	for (x = 0; x < 1; ++x) {
		CUDA_SAFE_CALL(cudaGLUnmapBufferObject(pbo[x]));
	}

	times[2] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

    // display results
    glClear(GL_COLOR_BUFFER_BIT);

    // draw image from PBO
    glDisable(GL_DEPTH_TEST);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo[0]);

	//Use this to expand/contract pixel area
	//glPixelZoom(2, 2);

	//we could do a glEnable(GL_BLEND) to use source/destination alpha
	//if we actually used multiple pixel unpack buffers
    
	//Color indexing works, but is ridiculously slow.
	//glDrawPixels(width, height, GL_COLOR_INDEX, GL_UNSIGNED_INT, 0);
    glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, 0);
	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

    glutSwapBuffers();
    glutReportErrors();

	times[3] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

    CUT_SAFE_CALL(cutDeleteTimer(hTimer));

	return times;
}

float* CPUFrame(int iterations, unsigned int hTimer) {
	double x,y,xn,yn;
	int xt,yt;
	int i;
	Param curp;
	ParamList* cp = ParamTween_cur(ptween);
	ParamTween_increment(ptween);
	
	float* times = (float*) malloc(sizeof(float)*4);
    CUT_SAFE_CALL(cutResetTimer(hTimer));
    CUT_SAFE_CALL(cutStartTimer(hTimer));

	glClear(GL_COLOR_BUFFER_BIT);
    glDisable(GL_DEPTH_TEST);

	memset(cpudata, 0, sizeof(int)*width*height);

	//The initialization time
	times[0] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

	x = RANDDBL(cp->xmin, cp->xmax);
	y = RANDDBL(cp->ymin, cp->ymax);

	for (i = 0; i < iterations; ++i) {
		if (RAND1() < 0.5) {
			curp = cp->paramarr[0];
		} else {
			curp = cp->paramarr[1];
		}

		xn = curp.a * x + curp.c * y + curp.e;
		yn = curp.b * x + curp.d * y + curp.f;

		xt = (int) ((xn - cp->xmin) / (cp->xmax - cp->xmin) * width);
		yt = (int) ((yn - cp->ymin) / (cp->ymax - cp->ymin) * height);

		if (xt < width && xt > 0 && yt < height && yt > 0) {
			cpudata[yt*width + xt] += 0x01;
		}
		x = xn;
		y = yn;
	}

	//Drawing time
	times[1] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

	for (xt = 0; xt < width; ++xt) {
		for (yt = 0; yt < height; ++yt) {
			cpudata[yt*width+xt] = (int)(sqrt(sqrt((float) cpudata[yt*width+xt] / 256.0))*256.0);
		}
	}
	
	//Mapping time
	times[2] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

    glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, cpudata);
    glutSwapBuffers();
    glutReportErrors();

	//Finishing time
	times[3] = cutGetTimerValue(hTimer);
	CUT_SAFE_CALL(cutResetTimer(hTimer));

	return times;
}

void displayF() {
    char title[256];
    sprintf(title, "Fractal %s", "Display");  
    glutSetWindowTitle(title);

    unsigned int hTimer;
    CUT_SAFE_CALL(cutCreateTimer(&hTimer));

	float* times = GPUFrame(1024*300, hTimer);

	printf("Init/Draw/Map/Finish: %6f/%6f/%6f/%6f\n", times[0], times[1], times[2], times[3]);

	free(times);
}

void idleF() {
	//do while idle...
	glutPostRedisplay();
}

void cleanup() {
	int x;
	for (x = 0; x < 1; ++x) {
		CUDA_SAFE_CALL(cudaGLUnregisterBufferObject(pbo[x]));
	}
	glDeleteBuffersARB(BUFFERLEN, pbo);

	//clean up the parameter storage space
	CUDA_SAFE_CALL(cudaFree(devparams));
	free(hostparams);
}

void startGLUT() {
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(width, height);
	glutCreateWindow("GPU Computation to Direct Display");
    glutDisplayFunc(displayF);
	glutIdleFunc(idleF);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0 GL_ARB_pixel_buffer_object")) {
        fprintf(stderr, "Required OpenGL pixel buffer extension missing.");
        exit(-1);
    }
}

void createBuffers() {
	int x;
	
	//generate buffer identifiers
	glGenBuffersARB(BUFFERLEN, pbo);

	// create pixel buffer objects
	for (x = 0; x < 1; ++x) {
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo[x]);
		glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, width*height*sizeof(GLubyte)*4, 0, GL_DYNAMIC_DRAW_ARB);
		glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);
		CUDA_SAFE_CALL(cudaGLRegisterBufferObject(pbo[x]));
	}
}

void createMap() {
	GLfloat* mapd = (GLfloat*) malloc(sizeof(GLfloat)*COLORMAPSIZE);
	int x;
	//a value between 0 and 1
	float pct;

	for (x = 0; x < COLORMAPSIZE; ++x) {
		pct = ((float) x) / ((float) COLORMAPSIZE);
		mapd[x] = sqrt(sqrt(pct));
	}

	glPixelMapfv(GL_PIXEL_MAP_I_TO_R, 256, mapd);
}

void loadRandom() {
	int data[RANDNODIM];
	float fldata[RANDFLOATDIM];
	int x;
	int y;

	//clear the array we're passing over.
	memset(data, 0, sizeof(data));

	//fill the array with random bits
	for (x = 0; x < RANDNODIM; ++x) {
		for (y = 0; y < sizeof(int)*8; ++y) {
			data[x] |= ((RAND1() > 0.5) ? 1 : 0) << y;
		}
	}

	//put some random numbers in the random float array
	for (x = 0; x < RANDFLOATDIM; ++x) {
		fldata[x] = (float) RAND1();
	}

	//copy random numbers to the device
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(randnos,data,sizeof(data)));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(randfloats,fldata,sizeof(fldata)));
}

void loadParameters(StringList filenames) {
	//Create parameter list support
	CUDA_SAFE_CALL(cudaMalloc((void**)&devparams, sizeof(GPUParams)*BUFFERLEN));
	hostparams = (GPUParams*) malloc(sizeof(GPUParams)*BUFFERLEN);

	//For a single parameter list
	/*
	plst = ParamList_init();
	ParamList_load(plst, 
		"C:\\Documents and Settings\\mrcaps\\Desktop\\Proj\\fractal\\ifs\\a001.ifs.txt");
	GPUParams_print(getGPUParams(plst));
	*/

	//For multiple parameter lists
	ptween = ParamTween_init();
	ParamTween_load(ptween, filenames.strings, filenames.len, 200);
}

int main(int argc, char **argv){
	CUT_DEVICE_INIT(argc, argv);

	glutInit(&argc, argv);
	startGLUT();
    createBuffers();

	//create the CPU data array
	cpudata = (int*) malloc(sizeof(int)*width*height);

	loadRandom();

	StringList filenames;
	char* startup = getStartupDir(argv[0]);
	filenames = getFilenamesFromFile(startup, "list.txt");
	loadParameters(filenames);

	atexit(cleanup);

	//Load the color map. This is too slow.
	//createMap();

	glutMainLoop();

    CUT_EXIT(argc, argv);
}
