/*
Compute data on GPU and output directly to display
Ilari Shafer, Computer Architecture, Spring 2008
Portions from simpleTexture3D demo as noted below.
*/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <cutil.h>
#include <math.h>
#include <GL/glew.h>
#include <GL/glut.h>
#include <cuda_gl_interop.h>
#include <multithreading.h>

#include "util.h"

typedef unsigned int uint;

//the number of repetitions to run
const int REPS = 10000*8;
const int RANDNODIM = REPS/8;

const int width = 512, height = 512;

//the dimensions of the block size can't multiply to be more than 512.
const dim3 blockSize(16, 16, 1);
const dim3 gridSize(width / blockSize.x, height / blockSize.y);

//OpenGL pixel buffer
GLuint pbo;

//Maintain a string of random numbers in constant memory cache
//It's not as optimal as using texture memory, but it's much cleaner,
//and we don't need 2D access optimization.
__device__ __constant__ int randnos[RANDNODIM];

__global__ void kdraw(uint* out, int width, int height) {
	int x = blockIdx.x * blockDim.x + threadIdx.x;
	int y = blockIdx.y * blockDim.y + threadIdx.y;

	out[y*width + x] = randnos[((int) (x+y)/10)%3];
}

void render() {
	uint* output;
	CUDA_SAFE_CALL(cudaGLMapBufferObject((void**)&output, pbo));

    // call CUDA kernel, writing results to PBO
    kdraw<<<gridSize, blockSize>>>(output, width, height);
    CUT_CHECK_ERROR("kernel failed");

    CUDA_SAFE_CALL(cudaGLUnmapBufferObject(pbo));
}

void displayF() {
	//from NVidia demo simpleTexture3D.cu

	render();

    // display results
    glClear(GL_COLOR_BUFFER_BIT);

    // draw image from PBO
    glDisable(GL_DEPTH_TEST);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
    glDrawPixels(width, height, GL_RGBA, GL_UNSIGNED_BYTE, 0);
    glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

    glutSwapBuffers();
    glutReportErrors();
}

void idleF() {
	//do while idle...
	glutPostRedisplay();
}

void cleanup() {
	CUDA_SAFE_CALL(cudaGLUnregisterBufferObject(pbo));    
	glDeleteBuffersARB(1, &pbo);
}

void initOpenGL() {
    // create pixel buffer object
    // from NVidia demo simpleTexture3D.cu
    glGenBuffersARB(1, &pbo);
	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, pbo);
	glBufferDataARB(GL_PIXEL_UNPACK_BUFFER_ARB, width*height*sizeof(GLubyte)*4, 0, GL_STREAM_DRAW_ARB);
	glBindBufferARB(GL_PIXEL_UNPACK_BUFFER_ARB, 0);

	CUDA_SAFE_CALL(cudaGLRegisterBufferObject(pbo));
}

void loadRandom() {
	int data[RANDNODIM];
	int x;
	int y;

	//clear the array we're passing over.
	memset(data, 0, sizeof(data));

	//fill the array with random bits
	for (x = 0; x < RANDNODIM; ++x) {
		for (y = 0; y < sizeof(int)*8; ++y) {
			data[x] |= ((RAND1() > 0.5) ? 1 : 0) << y;
		}
	}

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(randnos,data,sizeof(data)));
}

int main(int argc, char **argv){
	CUT_DEVICE_INIT(argc, argv);

	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE);
	glutInitWindowSize(width, height);
	glutCreateWindow("GPU Computation to Direct Display");
    glutDisplayFunc(displayF);
	glutIdleFunc(idleF);

    glewInit();
    if (!glewIsSupported("GL_VERSION_2_0 GL_ARB_pixel_buffer_object")) {
        fprintf(stderr, "Required OpenGL pixel buffer extension missing.");
        exit(-1);
    }
    initOpenGL();

	loadRandom();

	atexit(cleanup);

	glutMainLoop();

    CUT_EXIT(argc, argv);
}
